package it.sauronsoftware.cron4j;

import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;

/**
 * <p>
 * A predictor is able to predict when a scheduling pattern will be matched.
 * </p>
 * <p>
 * Suppose you want to know when the scheduler will execute a task scheduled
 * with the pattern <em>0 3 * jan-jun,sep-dec mon-fri</em>. You can predict
 * the next <em>n</em> execution of the task using a Predictor instance:
 * </p>
 * 
 * <pre>
 * String pattern = &quot;0 3 * jan-jun,sep-dec mon-fri&quot;;
 * Predictor p = new Predictor(pattern);
 * for (int i = 0; i &lt; n; i++) {
 * 	System.out.println(p.nextMatchingDate());
 * }
 * </pre>
 * 
 * @author Carlo Pelliccia
 * @since 1.1
 */
public class Predictor {

	/**
	 * The scheduling pattern on which the predictor works.
	 */
	private SchedulingPattern schedulingPattern;

	/**
	 * The start time for the next prediction.
	 */
	private long time;

	/**
	 * It builds a predictor with the given scheduling pattern and start time.
	 * 
	 * @param schedulingPattern
	 *            The pattern on which the prediction will be based.
	 * @param start
	 *            The start time of the prediction.
	 * @throws InvalidPatternException
	 *             In the given scheduling pattern isn't valid.
	 */
	public Predictor(String schedulingPattern, long start)
			throws InvalidPatternException {
		this.schedulingPattern = new SchedulingPattern(schedulingPattern);
		this.time = (start / (1000 * 60)) * 1000 * 60;
	}

	/**
	 * It builds a predictor with the given scheduling pattern and start time.
	 * 
	 * @param schedulingPattern
	 *            The pattern on which the prediction will be based.
	 * @param start
	 *            The start time of the prediction.
	 * @throws InvalidPatternException
	 *             In the given scheduling pattern isn't valid.
	 */
	public Predictor(String schedulingPattern, Date start)
			throws InvalidPatternException {
		this(schedulingPattern, start.getTime());
	}

	/**
	 * It builds a predictor with the given scheduling pattern and the current
	 * system time as the prediction start time.
	 * 
	 * @param schedulingPattern
	 *            The pattern on which the prediction will be based.
	 * @throws InvalidPatternException
	 *             In the given scheduling pattern isn't valid.
	 */
	public Predictor(String schedulingPattern) throws InvalidPatternException {
		this(schedulingPattern, System.currentTimeMillis());
	}

	/**
	 * It returns the next matching moment as a millis value.
	 * 
	 * @return The next matching moment as a millis value.
	 */
	public synchronized long nextMatchingTime() {
		// Go a minute ahead.
		time += 60000;
		// Is it matching?
		if (schedulingPattern.match(time)) {
			return time;
		}
		// Ok, split the time!
		Calendar c = new GregorianCalendar();
		c.setTimeInMillis(time);
		int minute = c.get(Calendar.MINUTE);
		int hour = c.get(Calendar.HOUR_OF_DAY);
		int dayOfMonth = c.get(Calendar.DAY_OF_MONTH);
		int month = c.get(Calendar.MONTH);
		int year = c.get(Calendar.YEAR);
		// Gets the matchers.
		ValueMatcher minuteMatcher = schedulingPattern.minuteMatcher;
		ValueMatcher hourMatcher = schedulingPattern.hourMatcher;
		ValueMatcher dayOfMonthMatcher = schedulingPattern.dayOfMonthMatcher;
		ValueMatcher dayOfWeekMatcher = schedulingPattern.dayOfWeekMatcher;
		ValueMatcher monthMatcher = schedulingPattern.monthMatcher;
		// Let's find the next matching minute.
		do {
			if (minute > 59) {
				minute = 0;
				hour++;
			}
			if (minuteMatcher.match(minute)) {
				break;
			}
			minute++;
		} while (true);
		// The next matching hour.
		do {
			if (hour > 23) {
				hour = 0;
				dayOfMonth++;
			}
			if (hourMatcher.match(hour)) {
				break;
			}
			hour++;
		} while (true);
		do {
			// The next matching dayOfMonth.
			do {
				if (dayOfMonth > 31) {
					dayOfMonth = 1;
					month++;
				}
				if (dayOfMonthMatcher.match(dayOfMonth)) {
					break;
				}
				dayOfMonth++;
			} while (true);
			// The next matching month.
			do {
				if (month > Calendar.DECEMBER) {
					month = Calendar.JANUARY;
					year++;
				}
				if (monthMatcher.match(month + 1)) {
					break;
				}
				month++;
			} while (true);
			// Is this ok?
			c = new GregorianCalendar();
			c.set(Calendar.MINUTE, minute);
			c.set(Calendar.HOUR_OF_DAY, hour);
			c.set(Calendar.DAY_OF_MONTH, dayOfMonth);
			c.set(Calendar.MONTH, month);
			c.set(Calendar.YEAR, year);
			// Day-of-month/month compatibility check.
			int oldDayOfMonth = dayOfMonth;
			int oldMonth = month;
			dayOfMonth = c.get(Calendar.DAY_OF_MONTH);
			month = c.get(Calendar.MONTH);
			if (month != oldMonth || dayOfMonth != oldDayOfMonth) {
				// Do it again.
				continue;
			}
			// Day-of-week compatibility check.
			int dayOfWeek = c.get(Calendar.DAY_OF_WEEK);
			if (!dayOfWeekMatcher.match(dayOfWeek - 1)) {
				// Do it again...
				dayOfMonth++;
				continue;
			}
			// Seems it matchs!
			this.time = (c.getTimeInMillis() / (1000 * 60)) * 1000 * 60;
			break;
		} while (true);
		// Here it is.
		return this.time;
	}

	/**
	 * It returns the next matching moment as a {@link Date} object.
	 * 
	 * @return The next matching moment as a {@link Date} object.
	 */
	public synchronized Date nextMatchingDate() {
		return new Date(nextMatchingTime());
	}

}
